from dataclasses import dataclass
import logging
import regex
from phantombuster.plumbing import calculate_threshold
from phantombuster.io_ import read_file

VALID_TAGS = ['B2', 'query', 'BC', 'name', 'seq']


@dataclass
class InputFile:
    path: str
    group: str
    prefix: str

    def _to_json_e(self):
        return {'path': self.path, 'group': self.group, 'prefix': self.prefix}

    @classmethod
    def _from_json_e(cls, d: dict):
        return InputFile(**d)


@dataclass
class InputGroup:
    files: list[InputFile]
    group: str

    def _to_json_e(self) -> dict:
        return {'files': [f._to_json_e() for f in self.files], 'group': self.group}

    @classmethod
    def _from_json_e(cls, d: dict) -> 'InputGroup':
        ig = InputGroup(files=[InputFile._from_json_e(f) for f in d['files']], group=d['group'])
        return ig


class RegexDictionary:

    def __init__(self, d=None):
        if d is None:
            d = {}
        self._groups = d

    def __eq__(self, o):
        if isinstance(o, RegexDictionary):
            return self._groups == o._groups
        else:
            return NotImplemented()

    def add_regex(self, tag, rex, prefix='', group='*'):
        d = self._groups.get(group, {})
        d[prefix+tag] = rex
        self._groups[group] = d

    def get_regexes_for_group(self, group):
        d1 = self._groups.get(group, {})
        d2 = self._groups.get('*', {})
        d = d1|d2
        assert len(d) > 0
        return {key: regex.compile(re) for key, re in d.items()}

    def _to_json_e(self):
        return self._groups

    @classmethod
    def _from_json_e(cls, d):
        r = RegexDictionary(d)
        return r


def read_regex_file(path):
    table = read_file(path)
    column_names = ['group', 'prefix', 'tag', 'regex']

    tags = table['tag'].to_list()

    invalid_tags = [tag for tag in tags if tag not in VALID_TAGS]
    if len(invalid_tags) > 0:
        raise KeyError(f'Invalid tags: {", ".join(invalid_tags)}')

    regexes = table['regex'].to_list()

    if 'group' not in table.columns:
        groups = ['*'] * len(tags)
    else:
        groups = table['group'].to_list()

    if 'prefix' not in table.columns:
        prefixs = [''] * len(tags)
    else:
        prefixs = table['prefix'].to_list()

    regex_dict = RegexDictionary()

    for group, tag, re, prefix in zip(groups, tags, regexes, prefixs):
        regex_dict.add_regex(tag, re, prefix, group)
    return regex_dict


def read_barcode_hierarchy_file(name):
    table = read_file(name)
    column_names = ["barcode", "type", "referencefile", "threshold", "min_length", "max_length"]
    missing_columns = [col for col in column_names if col not in table.columns]
    if len(missing_columns) > 0 :
        raise Exception(f"Barcode Hierarchy File Incorrect, missing columns: {missing_columns}") # TODO better error reporting

    def parse_length(value):
        if value == "-":
            return None
        else:
            try:
                return int(value)
            except Exception:
                raise Exception("Can not parse min or max length correctly in barcode hierarchy file, must be either '-' or an integer")

    def parse_type(value):
        if value == "reference":
            return "reference"
        else:
            return "random"

    barcodes = [{"name": name, "type": parse_type(type), "referencefile": reference, "threshold": threshold,
                 "min_length": parse_length(min_length), "max_length": parse_length(max_length)}
                 for name, type, reference, threshold, min_length, max_length in zip(*[table[name].to_list() for name in column_names])]

    for bc in barcodes:
        if bc["type"] == "reference":
            table = read_file(bc["referencefile"])
            bc["reference"] = {bc: name for bc, name in zip([str(bc) for bc in table["barcode"].to_list()], [str(name) for name in table["name"].to_list()])}
            if bc["threshold"] == "auto":
                bc["threshold"] = calculate_threshold(bc["reference"])
            else:
                try:
                    bc["threshold"] = int(bc["threshold"])
                except Exception:
                    raise Exception("Can not properly read threshold column of barcode hierarchy file, needs to be 'auto' or an integer")
    return barcodes


def read_input_files_file(path):
    table = read_file(path)

    column_names = ['file', 'group', 'prefix']

    files = table['file'].to_list()

    if 'group' not in table.columns:
        groups = table["file"].to_list()
    else:
        groups = table['group'].to_list()

    if 'prefix' not in table.columns:
        prefixs = [''] * len(files)
    else:
        prefixs = table['prefix'].to_list()

    input_files = [InputFile(file, group if group is not None else "", prefix if prefix is not None else "") for file, group, prefix in zip(files, groups, prefixs)]
    groups = {}
    for f in input_files:
        l = groups.get(f.group, [])
        l.append(f)
        groups[f.group] = l

    igs = []
    for group_name, files in groups.items():
        igs.append(InputGroup(files, group_name))

    return igs



import sqlite3
import dataclasses
from datetime import datetime

import typing
from typing import Set, Type, Any, TypeVar, List, Literal, Union, Dict, NoReturn, Optional
from types import TracebackType

Q = TypeVar("Q")

StoreID = typing.NewType("StoreID", int)

Operator = Literal['<', '<=', '=', '!=', '>=', '>', 'and', 'or']
JoinHow = Literal['inner', 'left']
Order = Literal['asc', 'desc']


def assert_never(value: NoReturn) -> NoReturn:
    assert False, f'Unhandled value: {value} ({type(value).__name__})'


class Column:

    def __init__(self, name: str):
        self.name = name

    def __lt__(self, other: "Column") -> 'Condition':
        return Condition(self, other, "<")

    def __le__(self, other: "Column") -> 'Condition':
        return Condition(self, other, "<=")

    def __eq__(self, other: Union["Column", Any]) -> Union[bool, 'Condition']: #type: ignore
        return Condition(self, other, "=")

    def __ne__(self, other: Union["Column", Any]) -> Union[bool, 'Condition']: #type: ignore
        return Condition(self, other, "!=")

    def __ge__(self, other: "Column") -> 'Condition':
        return Condition(self, other, ">=")

    def __gt__(self, other: "Column") -> 'Condition':
        return Condition(self, other, ">")


class Condition:

    def __init__(self, c1: Column, c2: Union[Any, Column], operator: Operator):
        self.c1 = c1
        self.c2 = c2
        self.operator = operator

    def __and__(self, other: "Condition") -> 'AndCondition':
        return AndCondition([self, other])

    def __or__(self, other: "Condition") -> 'OrCondition':
        return OrCondition([self, other])

    def __invert__(self) -> 'Condition':
        return ICondition(self)

    def __str__(self) -> str:
        return f"Condition({self.c1} {self.operator} {self.c2})"


class AndCondition(Condition):

    def __init__(self, cs: List[Condition]):
        self.cs = cs

    def __and__(self, other: Condition) -> 'AndCondition':
        return AndCondition(self.cs+[other])


class OrCondition(Condition):

    def __init__(self, cs: List[Condition]):
        self.cs = cs

    def __or__(self, other: Condition) -> 'OrCondition':
        return OrCondition(self.cs+[other])


class ICondition(Condition):

    def __init__(self, c: Condition):
        self.c = c

    def __invert__(self) -> Condition:
        return self.c


class Table:

    def __init__(self, T: type[Q]):
        self.T = T
        self._name = T.__name__

    @property
    def name(self) -> str:
        return self._name

    def join(self, other: 'Table', on: Column, how: JoinHow="inner") -> 'Join':
        return Join(self, other, on, on, how)

    def where(self, condition: Condition) -> 'ConditionedTable':
        return ConditionedTable(self, condition)

    def order_by(self, col: Column, order: Order="asc") -> 'OrderedTable':
        return OrderedTable(self, col, order)


class Join(Table):

    def __init__(self, t1: Table, t2: Table, on1: Column, on2: Column, how: JoinHow):
        self.t1 = t1
        self.t2 = t2
        self.on1 = on1
        self.on2 = on2
        self.how = how

    @property
    def name(self) -> str:
        return self.t2.name

    @property
    def inner_name(self) -> str:
        return self.t1.name

class ConditionedTable(Table):

    def __init__(self, table: Table, condition: Condition):
        self.table = table
        self.condition = condition

    @property
    def name(self) -> str:
        return self.table.name

    def where(self, condition: Condition) -> 'ConditionedTable':
        return ConditionedTable(self.table, self.condition & condition)


class OrderedTable(Table):

    def __init__(self, table: Table, column: Column, order: Order):
        self.table = table
        self.column = column
        self.order = order

    @property
    def name(self) -> str:
        return self.table.name


col = Column
table = Table



class Store:

    def __init__(self, path: str):
        self._path = path
        self._tables: Set[str] = set()
        self._types: Set[type] = set()

        self._initialized = False
        self._in_transaction = 0

    def __enter__(self) -> "Store":
        self._begin()
        return self

    def __exit__(self, exc_type: type[BaseException], exc_value: BaseException, exc_traceback: TracebackType) -> None:
        if self._in_transaction:
            if exc_type is None:
                self._commit()
            else:
                self._rollback()

    def _register_dataclass(self, T: type) -> None:
        self._types.add(T)

    def _get_primary(self, T: type[Q]) -> str:
        for field in dataclasses.fields(T):
            if field.type == StoreID:
                return field.name
            if field.metadata.get("store.type", None) == "primary":
                return field.name
        return "rowid"

    def _create_table_query(self, T: type[Q]) -> str:
        create = "CREATE TABLE IF NOT EXISTS"
        tablename = T.__name__
        fields = []
        for field in dataclasses.fields(T):
            python_field_type = field.type

            optional = False
            if typing.get_origin(python_field_type) == typing.Union:
                a1, a2 = typing.get_args(python_field_type)
                if a1 is None:
                    a1, a2 = a2, a1
                python_field_type = a1
                optional = True

            field_name = field.name
            if python_field_type == str:
                field_type = "TEXT"
            elif python_field_type == bytes:
                field_type = "BLOB"
            elif python_field_type == datetime:
                field_type = "timestamp"
            elif python_field_type == int:
                field_type = "INTEGER"
            elif python_field_type == float:
                field_type = "REAL"
            elif python_field_type == bool:
                field_type = "BOOL"
            elif python_field_type == StoreID:
                field_type = "INTEGER PRIMARY KEY"
            else:
                raise Exception(f"Can not detect mapping of python type {python_field_type} to sqlite type")

            if optional and python_field_type != StoreID:
                field_type = field_type + " NOT NULL"

            if field.metadata.get("store.type", None) == "primary":
                if field.type != StoreID:
                    field_type = field_type + " PRIMARY KEY"
            fields.append(field_name + ' ' + field_type)
        field_statement = ", ".join(fields)
        return f'{create} {tablename}({field_statement});'

    def _begin(self) -> None:
        if self._in_transaction == 0:
            self._connection.execute("BEGIN TRANSACTION;")
        self._in_transaction += 1

    def _commit(self) -> None:
        self._in_transaction -= 1
        if self._in_transaction == 0:
            self._connection.execute("COMMIT;")

    def _rollback(self) -> None:
        self._connection.execute("ROLLBACK;")
        self._in_transaction = 0

    def create_db(self) -> None:
        sqlite3.register_converter("BOOL", lambda v: v == b"1")
        self._connection = sqlite3.connect(self._path, detect_types=sqlite3.PARSE_DECLTYPES, isolation_level=None)
        cur = self._connection.cursor()
        cur.execute("pragma journal_mode = WAL;")
        cur.execute("pragma synchronous = normal;")
        cur.execute("pragma temp_store = memory;")
        cur.execute("pragma mmap_size = 30000000000;")
        # cur.execute("pragma page_size = 32768;")

        with self:
            for dc in self._types:
                q = self._create_table_query(dc)
                cur.execute(q)
            for dc in self._types:
                self._tables.add(dc.__name__)

        self._initialized = True

    def create_table(self, dc: type[Q]) -> None:
        q = self._create_table_query(dc)
        cur = self._connection.cursor()
        with self:
            cur.execute(q)
        self._tables.add(dc.__name__)

    def insert(self, t: List[Q], update: bool=False) -> None:
        if not isinstance(t, list):
            t = [t]
        t0 = t[0]
        T = type(t0)
        tablename = T.__name__

        if not self._initialized:
            self.create_db()

        if tablename not in self._tables:
            self.create_table(T)

        aliases = [field.name for field in dataclasses.fields(type(t0))]
        placeholders = ["?" for field in dataclasses.fields(type(t0))]

        q = f"INSERT INTO {tablename} ({', '.join(aliases)}) VALUES ({', '.join(placeholders)})"

        if update:
            primary = self._get_primary(T)
            upsert_clause_l = [f" ON CONFLICT({primary}) DO UPDATE SET"]

            upsert_clause_l_fields = []
            for field in dataclasses.fields(T):
                upsert_clause_l_fields.append(f"{field.name}=?")
            upsert_clause_l.append(", ".join(upsert_clause_l_fields))
            upsert_clause = " ".join(upsert_clause_l)

            q += upsert_clause
            fillers = []
            for instance in t:
                tup = dataclasses.astuple(instance)
                fillers.append(tup+tup)
        else:
            fillers = [dataclasses.astuple(instance) for instance in t]

        q += ";"

        cur = self._connection.cursor()
        with self:
            cur.executemany(q, fillers)

    def get_single(self, T: Type[Q], **kwargs: Any) -> Optional[Q]:
        ts = self.get(T, **kwargs)
        if len(ts) == 1:
            return ts[0]
        elif len(ts) == 0:
            return None
        else:
            raise KeyError("Multiple values for query")

    def get(self, T: Type[Q], **kwargs: Any) -> List[Q]:
        fields = [field.name for field in dataclasses.fields(T)]
        tablename = T.__name__

        if tablename not in self._tables:
            return []

        if not self._initialized:
            self.create_db()

        where_clause = " and ".join([f"{key} = ?" for key in kwargs])
        if where_clause:
            where_clause = "WHERE " + where_clause
        else:
            where_clause = ''
        select_clause = f"SELECT {', '.join(fields)} from {tablename} "

        q = ' '.join([select_clause, where_clause, ';'])

        cur = self._connection.cursor()
        cur.execute(q, tuple([value for value in kwargs.values()]))
        rows = cur.fetchall()

        result = [T(*args) for args in rows]
        return result

    def _get_table_postfix(self, table_counter: Dict[str, int], name: str, increment: bool = True) -> str:
        table_count = table_counter.get(name, 0)
        if increment:
            table_counter[name] = table_count + 1
        if table_count == 0:
            table_postfix = ''
        else:
            table_postfix = str(table_count)
        return table_postfix

    def _get_value_as_str(self, tableid: str, value: Union[str, int, Column], param_container: List[Union[str, int]]) -> str:
        if isinstance(value, str) or isinstance(value, int) or isinstance(value, bytes) or isinstance(value, bool):
            param_container.append(value)
            return "?"
        elif isinstance(value, Column):
            return f"{tableid}.{value.name}"
        else:
            raise ValueError(f"Cannot deal with value of type {type(value)}")

    def _get_cond_str(self, tableid: str, cond: Condition, param_container: List[Union[str, int]]) -> str:
        if isinstance(cond, OrCondition):
            return "(" + " or ".join([self._get_cond_str(tableid, c, param_container) for c in cond.cs]) + ")"
        elif isinstance(cond, AndCondition):
            return "(" + " and ".join([self._get_cond_str(tableid, c, param_container) for c in cond.cs]) + ")"
        elif isinstance(cond, ICondition):
            return "NOT " + self._get_cond_str(tableid, cond.c, param_container)
        elif isinstance(cond, Condition):
            left = self._get_value_as_str(tableid, cond.c1, param_container)
            if cond.c2 is None:
                return f"{left} IS NULL"
            right = self._get_value_as_str(tableid, cond.c2, param_container)
            return f"{left} {cond.operator} {right}"
        else:
            assert_never(cond)

    def select(self, query: Union[Table, ConditionedTable, OrderedTable, Join], limit: Optional[int]=None) -> List[Q]:

        table_clauses: List[str] = []
        where_clauses: List[str] = []
        order_by = ""

        param_container: List[Union[str, int]] = []
        table_counter: Dict[str, int] = {}

        table: Optional[Table] = query

        while table is not None:
            if isinstance(table, OrderedTable):
                assert order_by == ""

                table_name = table.name
                table_postfix = self._get_table_postfix(table_counter, table_name, increment=False)
                table_id = table_name + table_postfix

                column = table.column.name if isinstance(table.column, Column) else table.column
                order_by = f'ORDER BY {table_id}.{column}' + (' ASC' if table.order == 'asc' else ' DESC')
                table = table.table
            elif isinstance(table, ConditionedTable):
                condition = table.condition
                table = table.table

                table_name = table.name
                table_postfix = self._get_table_postfix(table_counter, table_name, increment=False)
                table_id = table_name + table_postfix

                where_clauses.append(self._get_cond_str(table_id, condition, param_container))

            elif isinstance(table, Join):
                table_name = table.t2.name
                table_postfix = self._get_table_postfix(table_counter, table.t2.name)
                table_id = table_name + table_postfix

                inner_table_name = table.inner_name
                inner_table_postfix = self._get_table_postfix(table_counter, inner_table_name, increment=False)
                inner_table_id = inner_table_name + inner_table_postfix

                if table.how == 'inner':
                    join_op = 'INNER JOIN'
                elif table.how == 'left':
                    join_op = 'LEFT JOIN'
                else:
                    assert_never(table.how)

                table_clauses.insert(0, f'{join_op} {table_name} {table_id} ON {inner_table_id}.{table.on1} = {table_id}.{table.on2}')

                table = table.t1
            elif isinstance(table, Table):
                T = table.T

                tablename = T.__name__

                if tablename not in self._tables:
                    self.create_table(T)

                table_postfix = self._get_table_postfix(table_counter, tablename)
                table_id = tablename + table_postfix

                fields = [f'{table_id}.{field.name}' for field in dataclasses.fields(T)]

                table_clauses.insert(0, f"SELECT {', '.join(fields)} from {tablename}{table_postfix}")
                table = None
            else:
                assert_never(table)

        select_clause = " ".join(table_clauses)
        where_clause = ("WHERE " + " and ".join(where_clauses) ) if where_clauses else ''

        limit_clause = f'LIMIT {limit}' if limit else ''

        clauses = [select_clause, where_clause, order_by, limit_clause]

        q = ' '.join([clause for clause in clauses if clause])

        cur = self._connection.cursor()
        cur.execute(q, param_container)
        rows = cur.fetchall()

        result = [T(*args) for args in rows]
        return result







from typing import Optional
import os.path
import logging
import pysam
import gzip
from dataclasses import dataclass
import polars as pl

from phantombuster.exceptions import PhantomBusterUnknownFileType


STAGES = ["deduplication", "error_corrected", "hopping_removal", "thresholded"]

@dataclass
class PathsAndFiles:
    outdir: str
    indexdir: str
    summarydir: str
    datadir: str
    tmpdir: str
    prefix: Optional[str]
    bamfiles: Optional[list[str]]
    tsv_filename: Optional[str]
    sort_filename: Optional[str]
    rest_filename: Optional[str]
    grna_filename: Optional[str]
    indexfile: Optional[str]
    task_db_filename: Optional[str]

    def __init__(self, outdir: str, prefix: Optional[str], bamfiles: Optional[list[str]]) -> None:

        assert prefix not in ["index", "summary", "data", "plot", "tmp"]

        self.outdir = outdir
        self.indexdir = os.path.join(self.outdir, "index")
        self.summarydir = os.path.join(self.outdir, "summary")
        self.datadir = os.path.join(self.outdir, "data")
        self.plotdir = os.path.join(self.outdir, "plot")
        self.tmpdir = os.path.join(self.outdir, "tmp")

        self.prefix = prefix
        self.bamfiles = bamfiles

        if self.prefix:
            self.tsv_filename = self.prefixs + ".tsv"
        else:
            if self.bamfiles:
                self.tsv_filename = self.bamfiles[0].replace(".bam", ".tsv")
            else:
                self.tsv_filename = None

        if self.tsv_filename is not None:
            self.indexfile = self.tsv_filename.replace(".tsv", ".index")
            self.sort_filename = self.tsv_filename.replace(".tsv", ".sort.tsv.gz")
            self.rest_filename = self.tsv_filename.replace(".tsv", ".rest.tsv.gz")
            self.grna_filename = self.tsv_filename.replace(".tsv", ".grna.tsv.gz")
        else:
            self.indexfile = None
            self.sort_filename = None
            self.rest_filename = None
            self.grna_filename = None

        self.task_db_filename = os.path.join(self.tmpdir, ".tasks.sqlite3")

    @property
    def prefixs(self):
        if self.prefix is None:
            return ""
        else:
            return self.prefix

    def in_dir(self, filename):
        path = [self.outdir]
        if self.prefix:
            path.append(self.prefix)
        path.apend(filename)
        return os.path.join(*path)

    def in_data(self, filename):
        path = [self.datadir]
        if self.prefix:
            path.append(self.prefix)
        path.apend(filename)
        return os.path.join(*path)

    def stage_path(self, stage):
        if stage not in STAGES:
            raise Exception(f"Invalid stage {stage}, valid stages are {STAGES}")
        path = [self.datadir]
        if self.prefix:
            path.append(self.prefix)
        path.append(stage)
        return os.path.join(*path)

    def create(self) -> None:
        for d in [self.indexdir, self.summarydir, self.datadir, self.plotdir, self.tmpdir]:
            os.makedirs(d, exist_ok=True)
        for rel_path in STAGES:
            f = [self.datadir]
            if self.prefix:
                f.append(self.prefix)
            f.append(rel_path)
            d = os.path.join(*f)
            os.makedirs(d, exist_ok=True)


def generate_directory_structure(dirs):
    # generating directory structure
    logging.info("Creating output directories")
    for direct in [
        dirs.outdir,
        dirs.datadir,
        dirs.indexdir,
        dirs.summarydir,
        dirs.plotdir,
        dirs.tmpdir,
    ]:
        logging.info(f"Creating directory {direct}.")
        # creating output directory if not already created
        if not os.path.exists(direct):
            os.mkdir(direct)
        else:
            logging.info("output directory already exists, no need to create it")


def open_sequencing_file(path, type=None):
    if type is None:
        if path.endswith('.bam'):
            type = 'bam'
        elif path.endswith('.sam'):
            type = 'sam'
        elif path.endswith('.fastq.gz'):
            type = 'fastq'
        else:
            raise ValueError(f'Cannot detect type for file {path}')

    if type == 'bam':
        return BamSequencingFile(path)
    elif type == 'sam':
        return SamSequencingFile(path)
    elif type == 'fastq':
        return FastqSequencingFile(path)
    else:
        raise ValueError(f'Unknown file type {type}')


class SequencingFile:

    def __init__(self, path):
        pass

    def seek(self, idx):
        pass

    def tell(self):
        pass

    def __iter__(self):
        pass


class SamSequencingFile(SequencingFile):

    def __init__(self, path):
        self.path = path
        self.file = pysam.Samfile(path, "r", check_sq=False)
        self._last_tell = None

    def seek(self, idx):
        self.file.seek(idx)
        self._last_tell = None

    def tell(self):
        return self.file.tell()

    def __iter__(self):
        while True:
            try:
                read = next(self.file)
            except StopIteration:
                break

            try:
                query = read.query_sequence
                data = {'query': query}
                for key, value in read.tags:
                    data[key] = value
                yield data
            except KeyError:
                pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.file.close()


class BamSequencingFile(SamSequencingFile):

    def __init__(self, path):
        self.path = path
        self.file = pysam.Samfile(path, "rb", check_sq=False)
        self._last_tell = None


class FastqSequencingFile(SequencingFile):

    def __init__(self, path):
        self.path = path
        self._last_tell = None
        if path.endswith('.gz'):
            self.file = gzip.open(path)
        else:
            self.file = open(path)

    def seek(self, idx):
        self.file.seek(idx)

    def tell(self):
        return self.file.tell()

    def __iter__(self):
        while True:
            name = self.file.readline().strip()
            if name == b'':
                break
            seq = self.file.readline().strip()
            _ = self.file.readline()
            qual = self.file.readline().strip()
            read = {'name': name.decode('utf8'), 'seq': seq.decode('utf8'), 'qual': qual.decode('utf8')}
            yield read

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.file.close()


def read_file(name: str):
    if name.endswith('tsv'):
        return read_tsv(name)
    elif name.endswith("parquet"):
        return read_parquet(name)
    elif name.endswith("csv"):
        return read_csv(name)
    else:
        raise PhantomBusterUnknownFileType(f"File {name} has unknown file type. Supported are .csv, .tsv and .parquet.")


def read_parquet(path):
    return pl.scan_parquet(path)


def read_csv(path):
    return pl.read_csv(path)


def read_tsv(path):
    return pl.read_csv(path, separator="\t")


def write_parquet(table, path):
    if isinstance(table, pl.DataFrame):
        table.write_parquet(path)
    elif isinstance(table, pl.LazyFrame):
        table.sink_parquet(path, engine="streaming")
    else:
        raise ValueError(f"Cannot write table of type {table.type} as parquet")


def write_csv(table, path):
    table.write_csv(path)


def write_tsv(table, path):
    table.write_csv(path, separator="\t")



import pickle
import logging
import sqlite3
from typing import Any, Callable, Dict, List, NoReturn, Optional, Set, Union, cast
import json
import uuid
import datetime

from .messages import StoreMsg, RemoterException
from .task import Task
from .persisters import persister_factory, ErrorPersister
from .async_ import async_gather
from .serialization import error_to_string, string_to_error, serialize_kwargs, deserialize_kwargs, DATETIME_FORMAT

def parse_time_from_sql(time):
    return datetime.datetime.strptime(time, "%Y-%m-%d %H:%M:%S.%f")

def serialize_time_to_sql(time):
    return time.strftime("%Y-%m-%d %H:%M:%S.%f")

# ==== Storing Metadata ====


class Store:
    """Store Metadata of task results"""

    def __init__(self, path: str, remotes: List["Store"] = None, name=None) -> None:
        if name is None:
            name = uuid.uuid4().hex
        self._name = name
        self._logger = logging.getLogger(f"remoter.store.{self._name}")
        self._path = path
        self._connection = sqlite3.connect(self._path)
        self._remotes = remotes if remotes else []
        self._waiting_for_tasks = {}
        self.init_db()

    def init_db(self) -> None:
        cur = self._connection.cursor()
        cur.execute(
            "CREATE TABLE IF NOT EXISTS tasks (task_hash BLOB NOT NULL, datetime TEXT NOT NULL, function_module TEXT, function_name TEXT NOT NULL, function_hash BLOB NOT NULL, kwargs BLOB NOT NULL, result BLOB NOT NULL, result_type TEXT NOT NULL, failed BOOLEAN NOT NULL, time REAL NOT NULL, PRIMARY KEY(task_hash));"
        )
        cur.execute(
            "CREATE TABLE IF NOT EXISTS functions (function_module BLOB, function_name BLOB NOT NULL, function_hash BLOB NOT NULL, function_code BLOB NOT NULL, datetime TEXT NOT NULL, PRIMARY KEY(function_hash));"
        )

    def _load_metadata(self, task):
        cur = self._connection.cursor()
        t = (task.calc_hash(),)
        sql = "SELECT datetime, function_module, function_name, function_hash, kwargs, result, result_type, failed, time from tasks WHERE task_hash = ?"
        r = cur.execute(sql, t).fetchone()
        cur.close()

        if r is None:
            raise RemoterException("No result in store")

        datetime_, function_module, function_name, function_hash, kwargs, result, result_type, failed, time = r
        datetime_, result, result_type, failed, time, kwargs = parse_time_from_sql(datetime_), pickle.loads(result), result_type, bool(failed), float(time), deserialize_kwargs(json.loads(kwargs), self)
        msg = StoreMsg(
            datetime_,
            function_module,
            function_name,
            function_hash,
            kwargs,
            result,
            task.calc_hash(),
            result_type,
            failed,
            time,
        )
        return msg

    def _load_metadata_one(self, task_hashs):
        cur = self._connection.cursor()

        VALUES = ", ".join("?" for _ in task_hashs)
        sql = f"SELECT task_hash, datetime, function_module, function_name, function_hash, kwargs, result, result_type, failed, time from tasks WHERE result IS NOT NULL and task_hash in ({VALUES}) LIMIT 1"
        r = cur.execute(sql, task_hashs).fetchone()
        cur.close()

        if r is None:
            raise RemoterException("No result in store")

        task_hash, datetime_, function_module, function_name, function_hash, kwargs, result, result_type, failed, time = r
        task_hash, datetime_, result, result_type, failed, time, kwargs = task_hash, parse_time_from_sql(datetime_), pickle.loads(result), result_type, bool(failed), float(time), deserialize_kwargs(json.loads(kwargs), self)
        msg = StoreMsg(
            datetime_,
            function_module,
            function_name,
            function_hash,
            kwargs,
            result,
            task_hash,
            result_type,
            failed,
            time,
        )
        return msg


    def _make_task_from_sqlite_return(self, r):
        task_hash, datetime_, function_module, function_name, function_hash, kwargs, result, result_type, failed, time = r
        datetime_, result, result_type, failed, time, kwargs = parse_time_from_sql(datetime_), pickle.loads(result), result_type, bool(failed), float(time), deserialize_kwargs(json.loads(kwargs), self)
        persister = persister_factory(result_type)
        task = Task(persister=persister, store=self, datetime=datetime_, function_module=function_module, function_name=function_name, function_hash=function_hash, kwargs=kwargs, result=result, task_hash=task_hash, done=True, failed=failed, time=time)
        return task

    def load_from_hash(self, hash) -> 'Task':
        cur = self._connection.cursor()
        t = (hash,)
        sql = "SELECT task_hash, datetime, function_module, function_name, function_hash, kwargs, result, result_type, failed, time from tasks WHERE task_hash = ?"
        r = cur.execute(sql, t).fetchone()
        cur.close()

        if r is None:
            task = Task(persister=persister_factory("remoter.InbasePersister()"), store=self, datetime=None, function_module=None, function_name=None, function_hash=None, kwargs=None, result=None, task_hash=hash, done=False, failed=False, time=None)
        else:
            task = self._make_task_from_sqlite_return(r)
        return task

    def _insert_from_hash(self, task, do_load=True) -> 'Task':
        cur = self._connection.cursor()
        t = (task.task_hash,)
        sql = "SELECT task_hash, datetime, function_module, function_name, function_hash, kwargs, result, result_type, failed, time from tasks WHERE task_hash = ?"
        r = cur.execute(sql, t).fetchone()
        cur.close()

        if r is not None:
            task_hash, datetime_, function_module, function_name, function_hash, kwargs, result, result_type, failed, time = r
            datetime_, result, result_type, failed, time, kwargs = parse_time_from_sql(datetime_), pickle.loads(result), result_type, bool(failed), float(time), deserialize_kwargs(json.loads(kwargs), self)
            persister = persister_factory(result_type)
            task.persister = persister
            task.datetime = datetime_
            task.function_module = function_module
            task.function_name = function_name
            task.function_hash = function_hash
            task.kwargs = kwargs
            task.result = result
            task.failed = failed
            task.time = time
            if do_load:
                task.load()
        return task

    def load(self, task) -> Any:
        o_task = task
        msg = self._load_metadata(task)
        persister = persister_factory(msg.result_type)
        self._logger.debug("For task %s using persister %s", task, persister)
        task.failed = msg.failed
        task = persister.from_msg(msg)
        task.store = self
        self._logger.debug("Persister returned task %s", task)
        for col in ["datetime", "function_module", "function_name", "function_hash", "kwargs", "result", "persister", "failed", "time"]:
            setattr(o_task, col, getattr(task, col))
        if msg.failed:
            o_task.failed = True
            o_task.exception = task.exception
        return task.result

    def load_all(self):
        cur = self._connection.cursor()
        sql = "SELECT task_hash, datetime, function_module, function_name, function_hash, kwargs, result, failed, time FROM tasks"
        rows = cur.execute(sql).fetchall()
        cur.close()
        tasks = [self._make_task_from_sqlite_return(r) for r in rows]
        return tasks

    def _income(self, msg):
        cur = self._connection.cursor()
        t = (
            msg.task_hash,
            serialize_time_to_sql(msg.datetime),
            msg.function_module,
            msg.function_name,
            msg.function_hash,
            json.dumps(serialize_kwargs(msg.kwargs)),
            pickle.dumps(msg.result),
            msg.result_type,
            msg.failed,
            msg.time,
        )
        try:
            cur.execute("INSERT INTO tasks VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", t)
        except sqlite3.IntegrityError:
            self._logger.debug("Tried to insert a result into the store which is already present")
        except sqlite3.InterfaceError as e:
            self._logger.error("InterfaceError, task is: %s.", t)
            raise e
        cur.close()
        self._connection.commit()

        if msg.task_hash in self._waiting_for_tasks:
            for task in self._waiting_for_tasks[msg.task_hash]:
                self._insert_from_hash(task, do_load=False)
            del self._waiting_for_tasks[msg.task_hash]

    def _funcsave(self, msg):
        cur = self._connection.cursor()
        t = (
            msg.function_module,
            msg.function_name,
            msg.function_hash,
            msg.function_code,
            serialize_time_to_sql(msg.datetime),
        )
        cur.execute("INSERT INTO functions VALUES (?, ?, ?, ?, ?)", t)
        cur.close()
        self._connection.commit()

    async def income(self, msg):
        self._income(msg)

    async def save(self, task) -> None:
        if task.failed:
            task.persister = ErrorPersister()
        msg = task.persister.to_msg(task)
        await self.income(msg)
        await async_gather(*(remote.income(msg) for remote in self._remotes))

    def _save_force(self, task) -> None:
        msg = task.persister.to_msg(task)
        self._income(msg)


class Remote(Store):
    """Forwards incoming StoreMsgs to remote stores

    Used by Workers to forward StoreMsgs to the Server"""

    def __init__(self, server, name=None):
        if name is None:
            name = uuid.uuid4().hex
        self._name = name
        self._logger = logging.getLogger(f"remoter.store.Remote.{self._name}")
        self._server = server

    async def save(self, task) -> None:
        raise NotImplemented

    def load(self, task) -> Any:
        raise NotImplemented

    async def income(self, msg) -> None:
        self._logger.debug("Forwarding")
        await self._server.send_async(msg)
        self._logger.debug("Forwarded")
        self._logger.debug("Receiving Forwarding Ack")
        await self._server.recv_async()
        self._logger.debug("Received Forwarding Ack")

